#!/usr/bin/php
<?

$rules = array(
	"Program" => array(
		array( "T_KW_PROGRAM", "T_IDENT", "T_SEMICOLON", "Blok", "T_DOT", "T_NONE" ) /* T_NONE == EOF */
	),
	"AditivniOperator" => array(
		array( "T_OP_MINUS" ),
		array( "T_OP_PLUS" )
	),
	"Blok" => array(
		array( "UsekDeklaraciPromennych", "SlozenyPrikaz" )
	),
	"CastElse" => array(
		array(),
		array( "T_KW_ELSE", "Prikaz" )
	),
	"CastFor" => array(
		array( "T_KW_DOWNTO" ),
		array( "T_KW_TO" )
	),
	"DeklaracePromennych" => array(
		array( "SeznamIdentifikatoru", "T_COLON", "OznaceniTypu" )
	),
	"Faktor" => array(
		array( "T_INT" ),
		array( "T_IDENT" ),
		array( "T_REAL" ),
		array( "Vyraz" )
	),
	"MultiplikativniOperator" => array(
		array( "T_OP_DIV_INT" ),
		array( "T_OP_MUL" ),
		array( "T_OP_DIV" ),
		array( "T_OP_MOD" )
	),
	"OznaceniTypu" => array(
		array( "T_KW_INT" ),
		array( "T_KW_REAL" )
	),
	"Podminka" => array(
		array( "Vyraz", "RelacniOperator", "Vyraz" )
	),
	"PosloupnostPrikazu" => array(
		array( "Prikaz", "ZbytekPosloupnostiPrikazu" )
	),
	"PrazdnyPrikaz" => array(
		array()
	),
	"PrikazFor" => array(
		array( "T_KW_FOR", "T_IDENT", "T_OP_ASSIGN", "Vyraz", "CastFor", "Vyraz", "T_KW_DO", "Prikaz" )
	),
	"PrikazIf" => array(
		array( "T_KW_IF", "Podminka", "T_KW_THEN", "Prikaz", "CastElse" )
	),
	"Prikaz" => array(
		array( "PrazdnyPrikaz" ),
		array( "PrikazFor" ),
		array( "PrikazIf" ),
		array( "PrikazWhile" ),
		array( "PrikazWriteln" ),
		array( "PrirazovaciPrikaz" ),
		array( "SlozenyPrikaz" )
	),
	"PrikazWhile" => array(
		array( "T_KW_WHILE", "Podminka", "T_KW_DO", "Prikaz" )
	),
	"PrikazWriteln" => array(
		array( "T_KW_WRITELN", "T_BRACKET_BEGIN", "Vyraz", "T_BRACKET_END" )
	),
	"PrirazovaciPrikaz" => array(
		array( "T_IDENT", "T_OP_ASSIGN", "Vyraz" )
	),
	"RelacniOperator" => array(
		array( "T_OP_LT" ),
		array( "T_OP_LE" ),
		array( "T_OP_NEQ" ),
		array( "T_OP_EQ" ),
		array( "T_OP_GT" ),
		array( "T_OP_GE" )
	),
	"SeznamIdentifikatoru" => array(
		array( "T_IDENT", "ZbytekSeznamuIdentifikatoru" )
	),
	"SlozenyPrikaz" => array(
		array( "T_KW_BEGIN", "PosloupnostPrikazu", "T_KW_END" )
	),
	"Term" => array(
		array( "Faktor", "ZbytekTermu" )
	),
	"UsekDeklaraciPromennych" => array( 
		array(),
		array( "T_KW_VAR", "DeklaracePromennych", "T_SEMICOLON", "ZbytekDeklaraciPromennych" )
	),
	"Vyraz" => array(
		array( "Znamenko", "Term", "ZbytekVyrazu" )
	),
	"ZbytekDeklaraciPromennych" => array(
		array(),
		array( "DeklaracePromennych", "T_SEMICOLON", "ZbytekDeklaraciPromennych" )
	),
	"ZbytekPosloupnostiPrikazu" => array(
		array(),
		array( "T_SEMICOLON", "Prikaz", "ZbytekPosloupnostiPrikazu" ),
	),
	"ZbytekSeznamuIdentifikatoru" => array(
		array(),
		array( "T_COMMA", "SeznamIdentifikatoru" )
	),
	"ZbytekTermu" => array(
		array(),
		array( "MultiplikativniOperator", "Faktor", "ZbytekTermu" )
	),
	"ZbytekVyrazu" => array(
		array( "AditivniOperator", "Term", "ZbytekVyrazu" ),
		array() 
	),
	"Znamenko" => array(
		array(),
		array( "T_OP_MINUS" ),
		array( "T_OP_PLUS" )
	)
);
// else patri vzdy k nejblizsimu if

$global_code = "/*** generated code ***/\n\n";
$global_prototypes = "/*** generated prototypes ***/\n\n";
$first = array();
$follow = array();

function is_terminal ($sym) {
	global $rules;
	return !isset($rules[$sym]);
}

function build_fn($src) {
	global $rules;
	global $global_code, $global_prototypes;
	global $first, $follow;

	$prototype = "static int synta_parse_$src(SYNTA *synta)";
	$global_prototypes .= $prototype.";\n";
	$code = $prototype."\n"
		."{\n"
		."\tTREE_ITER *iter_tmp;\n"
		."\tDBG_MSG(3, \"synta_parse_$src\");\n"
		."\n"
		."\ttree_update(synta->tree_iter, \"".$src."\", -1);\n"
		."\titer_tmp = tree_iter_dup(synta->tree_iter);\n"
		."\ttree_add_child(synta->tree_iter);\n"
		."\n"
		."\tswitch(synta->token_type) {\n"
	;

	$skip = array();
	
	foreach($rules[$src] as $rule) {
		$code .= "\t\t/* rule: $src => { ".join($rule, ", ")." } */\n";
		if (count($rule) == 0) {
			$code .= "\n";
			continue;
		}

		$cur_first = is_terminal($rule[0]) ? array($rule[0]) : $first[$rule[0]];
		$cur_follow = $follow[$rule[0]];

		if (empty($cur_first[0])) {
			$select = array_merge($cur_first, $cur_follow);
			sort($select);
			$select = array_slice(array_unique($select), 1);
		} else {
			$select = $cur_first;
		}

		$skip = array_merge($skip, $select);
		
		foreach ($select as $sym) {
			$code .= "\t\tcase $sym:\n";
		}

		foreach($rule as $sym) {
			if (is_terminal($sym)) {
				$code .= "\t\t\tsynta_read_next(synta, $sym);\n";
			} else {
				$code .= "\t\t\tsynta_parse_$sym(synta);\n";
			}
		}
		$code .= "\t\t\tbreak;\n\n";
	}
	if (count($follow[$src])) {
		$expand_to_nothing_is_empty = true;
		$code .= "\t\t/* expand to nothing (follow) */\n";
		sort($skip);
		foreach($follow[$src] as $sym) {
			if (!in_array($sym, $skip)) {
				$code .= "\t\tcase $sym:\n";
				$expand_to_nothing_is_empty = false;
			}
		}
		$code .= "\t\t\t/* skip: ".join(', ', $skip)." */\n";
		if ($expand_to_nothing_is_empty) {
			$code .= "\t\t\t/* empty */\n\n";
		} else {
			$code .= "\t\t\tbreak;\n\n";
		}
	}
	
	$code .= "\t\t/* error */\n"
		."\t\tdefault:\n"
		."\t\t\tfprintf(stderr, \"%s:%ld: Syntax error near \\\"%s\\\".\\n\", synta->lexa->file_name,\n"
		."\t\t\t\t\t\tlexa_get_line(synta->lexa), synta->token);\n"
		."\t\t\texit(-1);\n"
		."\t\t\tbreak;\n"
		."\t}\n"
		."\n"
		."\tif(tree_get_token(synta->tree_iter) == NULL)\n"
		."\t\ttree_delete(synta->tree_iter);\n"
		."\ttree_iter_destroy(synta->tree_iter);\n"
		."\tsynta->tree_iter = iter_tmp;\n"
		."\ttree_append(synta->tree_iter);\n"
		."\treturn 1;\n"
		."}\n\n";
	$global_code .= $code;
}

function calc_first() {
	global $rules;
	
	$first = array();
	$FIRST = array();
	
	foreach($rules as $sym => $x) {
		$first[$sym] = array();
		$FIRST[$sym] = array();
	}
	
	$changed = 1;
	$old_size = 0;
	while ($changed > 0) {
		$changed--;
		$size = 0;

		/* update $first */
		foreach ($first as $N => $f) {
			foreach($rules[$N] as $r) {
				if (isset($r[0]))
					$first[$N] = array_merge($first[$N], $r[0] ? (is_terminal($r[0]) ? array($r[0]) : $FIRST[$r[0]]) : array());
			}
			sort($first[$N]);
			$first[$N] = array_unique($first[$N]);
		}

		/* update $FIRST */
		foreach ($FIRST as $N => $f) {
			$FIRST[$N] = array();
			foreach($rules[$N] as $r) {
				if (empty($r[0])) {
					$FIRST[$N] = array_merge($FIRST[$N], array());
				} else if (is_terminal($r[0])) {
					$FIRST[$N] = array_merge($FIRST[$N], array($r[0]));
				} else if (!is_terminal($r[0]) && !empty($first[$r[0]][0])) {
					$FIRST[$N] = array_merge($FIRST[$N], $first[$r[0]]);
				} else if (!$first[$r[0]][0] && empty($first[$r[0]][0]) && isset($r[1])) {
					$FIRST[$N] = array_merge($FIRST[$N], array_slice($first[$r[0]], 1),
								is_terminal($r[1]) ? array($r[1]) : $FIRST[$r[1]]);
				}
			}
			sort($FIRST[$N]);
			$FIRST[$N] = array_unique($FIRST[$N]);
			$size += count($FIRST[$N]);
		}
		$changed = $size - $old_size;
		$old_size = $size;
	}

	return $FIRST;
}

function calc_follow($first, $rules) {
	$follow = array();

	foreach($rules as $sym => $rule) {
		$follow[$sym] = array();
	}

	$changed = 1;
	$old_size = 0;
	while ($changed > 0) {
		$size = 0;
		foreach ($follow as $sym => $sym_follow) {
			foreach($rules as $rule_sym => $rule) {
				foreach($rule as $r) {
					foreach($r as $i => $s) {
						if($s == $sym) {
							if (isset($r[$i + 1]) && !is_terminal($r[$i + 1])) {
								$follow[$sym] = array_merge($follow[$sym], $first[$r[$i + 1]], $follow[$r[$i + 1]]);
							}
							if (isset($r[$i + 1]) && is_terminal($r[$i + 1])) {
								array_push($follow[$sym], $r[$i + 1]);
							}
							if (!isset($r[$i + 1])) {
								$follow[$sym] = array_merge($follow[$sym], $follow[$rule_sym]);
							}
						}
					}
				}
			}
			sort($follow[$sym]);
			$follow[$sym] = array_unique($follow[$sym]);

			$size += count($follow[$sym]);
		}
		$changed = $size - $old_size;
		$old_size = $size;
	}
	
	foreach($rules as $sym => $rule) {
		foreach($rule as $r) {
			if (empty($r)) {
				if (empty($follow[$sym][0])) {
					$wanted_follow[$sym] = array_splice($follow[$sym], 1);
				} else {
					$wanted_follow[$sym] = $follow[$sym];
				}
			}
		}
	}
	
	return $wanted_follow;
}

function gen_tree($symbol, $indent = 1) {
	global $rules;

	if ($indent > 13)
		return;

	foreach ($rules[$symbol] as $id => $rule) {
		echo "  *".str_repeat("\t", $indent)."(".$id.") ".$symbol."\n";
		foreach ($rule as $r => $n) {
			if (is_terminal($n)) {
				echo "  *".str_repeat("\t", $indent + 1).$n."\n";
			} else {
				gen_tree($n, $indent + 1);
			}
		}
	}

}

/****************************************************************************/
echo "/***** content generated by gen-synta.php *****/\n\n";

echo " /* Rules:\n";

foreach($rules as $sym => $r) {
	foreach ($r as $n) {
		echo "  *\t".$sym." -> ".join(', ', $n)."\n";
	}
}

//echo "  */\n\n /* Derivacni strom pro vsechny validni vstupy:\n";
//gen_tree("Program");

echo "  */\n\n /* First:\n";
$first = calc_first();
foreach($first as $k => $v)
	echo "  *\t$k = { ".join($v, ", ")." }\n";
echo "  */\n\n";

/* blee, ta mrcha nechtela fungovat */
echo " /* Follow:\n";
$follow = calc_follow($first, $rules);
foreach($follow as $k => $v) {
	echo "  *\t$k = { ".join($v, ", ")." }\n";
}
echo "  */\n\n";

if (0) {
	build_fn("Program");
} else {
	foreach ($rules as $sym => $rule) {
		build_fn($sym);
	}
}

echo	$global_prototypes."\n\n"
	."/* Start */\n"
	."int synta_parse_start(SYNTA *synta)\n"
	."{\n"
	."	return synta_parse_Program(synta);\n"
	."}\n\n"
	.$global_code."\n\n";

echo "/***** end of generated content *****/\n/* vim:set filetype=c: */\n\n";
/****************************************************************************/
?>

