/*
 * lexa-table.h
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#ifndef JK_LEXA_TABLE_H
#define JK_LEXA_TABLE_H

/* transition from one state to anotrher */
typedef struct {
	int state;				/* actual state */
	unsigned char character;		/* characters for this transition */
	int next_state;				/* switch to this state if char_* match */
	int finish_type;			/* if non-zero, this state is terminating state (pattern matches) */
	int priority;				/* transition priority (only transition with the highest priority is used) */
} LEXA_TABLE_TRANS;

/* state transition table */
typedef struct {
	LEXA_TABLE_TRANS *tr;			/* array of transitions */
	int tr_size;				/* it's size */
	int tr_count;				/* number of transitions */
	int empty_state;			/* last state + 1 (empty state, probably first one) (autoincremented) */

	struct {
		unsigned ready :1;		/* array is ready for lookup */
	} flag;
} LEXA_TABLE;

/* table initialization
 */
LEXA_TABLE *lexa_table_init(int init_size, int empty_state_offset);

/* cleanup
 */
void lexa_table_celanup(LEXA_TABLE *table);

/* add transition into transition table
 *
 * returns transition if success, NULL if failed
 */
LEXA_TABLE_TRANS *lexa_table_add(LEXA_TABLE *table, int state, unsigned char character, int next_state, int finish_type, int priority);

/* lookup next state
 *
 * returns pointer to table of transitions, valid to next lexa_table_add() call only
 */
LEXA_TABLE_TRANS *lexa_table_lookup(LEXA_TABLE *table, int state, unsigned char ch);
LEXA_TABLE_TRANS *lexa_table_lookup_nosort(LEXA_TABLE *table, int state, unsigned char ch);


/* print state transition table
 */
void lexa_table_dump(LEXA_TABLE *lexa, FILE *file);


#endif

