/*
 * main.c
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "lexa.h"
#include "synta.h"

#define DEBUG 0
#include "debug.h"

/*** main *******************************************************************/

void print_mem_usage(void) {
	FILE *f;
	int c;

	fprintf(stderr, "\n\nProcess statistics:\n");
	f = fopen("/proc/self/status", "rt");
	if (f) {
		while ((c = fgetc(f)) != EOF)
			putc(c, stderr);
		fclose(f);
	} else {
		perror("print_mem_usage");
	}
}

int main(int argc, char *argv[]) {
	LEXA *la;
	SYNTA *sa;
	int type;
	char *token = NULL;
	
	/* Lexa: keyword list */
	struct {
		int tkw;	/* keyword type */
		char *kw;	/* keyword */
	} kw_list[] = {
		{ T_KW_BEGIN,	"begin"		},
		{ T_OP_DIV,	"div"		},
		{ T_KW_DO,	"do"		},
		{ T_KW_DOWNTO,	"downto"	},
		{ T_KW_ELSE,	"else"		},
		{ T_KW_END,	"end"		},
		{ T_KW_FOR,	"for"		},
		{ T_KW_IF,	"if"		},
		{ T_KW_INT,	"integer"	},
		{ T_OP_MOD,	"mod"		},
		{ T_KW_PROGRAM,	"program"	},
		{ T_KW_REAL,	"real"		},
		{ T_KW_THEN,	"then"		},
		{ T_KW_TO,	"to"		},
		{ T_KW_VAR,	"var"		},
		{ T_KW_WHILE,	"while"		},
		{ T_KW_WRITELN,	"writeln"	},
		{ 0,		NULL		}
	};
	int k;

	
#if DEBUG > 1
	atexit(print_mem_usage);
#endif

	DBG_MSG(1, "Initializing Lexa...");
	la = lexa_init(stdin, S_MAX);
	
	/* initialize */
	/* Lexa Table */
	DBG_MSG(2, "Lexa table...");
	/* 		state,     char_min, char_max,	char_exclude,	char_enum,	next_state,	finish_type */
	//lexa_add_tr(la, S_START,	  0, 0,		NULL,		0,		S_START,	T_NONE		); /* start position - added automaticaly */
	lexa_add_tr(la, S_START,	  0, 0,		NULL,		" \t\n\r",	S_WHITESPACE,	T_WHITESPACE	); /* white space */
	
	/* string -- quotes */
	/*
	lexa_add_tr(la, S_START,	  0, 0,		NULL,		"\"",		S_STRING_Q,	T_ERR		);
	lexa_add_tr(la, S_STRING_Q,	  0, 255,	"\\\"",		NULL,		S_STRING_Q,	T_ERR		);
	lexa_add_tr(la, S_STRING_Q,	  0, 0,		NULL,		"\\",		S_STRING_QS,	T_ERR		);
	lexa_add_tr(la, S_STRING_QS,	  0, 255,	NULL,		NULL,		S_STRING_Q,	T_ERR		);
	lexa_add_tr(la, S_STRING_Q,	  0, 0,		NULL,		"\"",		S_STRING_QT,	T_STRING	);
	*/
	
	/* integer */
	lexa_add_tr(la, S_START,	  0, 0,		NULL,		"0",		S_INTEGER_0,	T_INT		);
	lexa_add_tr(la, S_INTEGER_0,	  0, 0,		NULL,		"0",		S_INTEGER_0,	T_INT		);
	lexa_add_tr(la, S_START,	'1', '9',	NULL,		NULL,		S_INTEGER,	T_INT		);
	lexa_add_tr(la, S_INTEGER,	'0', '9',	NULL,		NULL,		S_INTEGER,	T_INT		);
	
	/* float */
	lexa_add_tr(la, S_START,	  0, 0,		NULL,		".",		S_REAL,		T_DOT		);
	lexa_add_tr(la, S_INTEGER,	  0, 0,		NULL,		".",		S_REAL,		T_REAL		);
	lexa_add_tr(la, S_INTEGER_0,	  0, 0,		NULL,		".",		S_REAL,		T_REAL		);
	lexa_add_tr(la, S_REAL,		'0', '9',	NULL,		NULL,		S_REAL,		T_REAL		);
	
	/* comments - C style */
	lexa_add_tr(la, S_START,	  0, 0,		NULL,		"/",		S_COMMENT_C_B,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_C_B,	  0, 0,		NULL,		"*",		S_COMMENT_C,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_C,	  0, 255,	"*",		NULL,		S_COMMENT_C,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_C,	  0, 0,		NULL,		"*",		S_COMMENT_C_E,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_C_E,	  0, 255,	"/",		NULL,		S_COMMENT_C,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_C_E,	  0, 0,		NULL,		"/",		S_COMMENT_C_T,	T_COMMENT	);
	
	/* comments - shell style */
	lexa_add_tr(la, S_START,	  0, 0,		NULL,		"#",		S_COMMENT_S,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_S,	  0, 255,	"\n",		NULL,		S_COMMENT_S,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_S,	  0, 0,		NULL,		"\n",		S_COMMENT_S_T,	T_COMMENT	);
	
	// comments - C++ style
	lexa_add_tr(la, S_COMMENT_C_B,	  0, 0,		NULL,		"/",		S_COMMENT_CC,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_CC,	  0, 255,	"\n",		NULL,		S_COMMENT_CC,	T_ERR		);
	lexa_add_tr(la, S_COMMENT_CC,	  0, 0,		NULL,		"\n",		S_COMMENT_CC_T,	T_COMMENT	);
	
	/* 		state,     char_min, char_max,	char_exclude,	char_enum,	next_state,	finish_type */
	
	/* keywords */
	lexa_add_tr(la, S_START,	  'a', 'z',	NULL,		NULL,		S_IDENT,	T_IDENT	);
	lexa_add_tr(la, S_START,	  'A', 'Z',	NULL, 		NULL,		S_IDENT,	T_IDENT	);
	lexa_add_tr(la, S_START,	    0, 0,	NULL,		"_",		S_IDENT,	T_IDENT	);
	
	/* identifikator */
	lexa_add_tr(la, S_IDENT,	  'a', 'z',	NULL,		NULL,		S_IDENT,	T_IDENT	);
	lexa_add_tr(la, S_IDENT,	  'A', 'Z',	NULL,		NULL,		S_IDENT,	T_IDENT	);
	lexa_add_tr(la, S_IDENT,	  '0', '9',	NULL,		NULL,		S_IDENT,	T_IDENT	);
	lexa_add_tr(la, S_IDENT,	    0, 0,	NULL,		"_",		S_IDENT,	T_IDENT	);
	for (k = 0; kw_list[k].kw; k++) {
		lexa_add_string_wf(la, kw_list[k].kw, kw_list[k].tkw, "qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890_", S_IDENT, T_IDENT);
		/*
		str[0] = kw_list[k].kw[0];
		lexa_add_tr(la, S_START,	    0, 0,	NULL,		str,	kw_list[k].sb,	T_IDENT	);
		for (i = 1; kw_list[k].kw[i]; i++) {
			str[0] = kw_list[k].kw[i];
			lexa_add_tr_p(la, kw_list[k].sb+i-1,    0, 0,	NULL,	str,	kw_list[k].sb+i,
					kw_list[k].kw[i+1] == 0 ? kw_list[k].tkw : T_IDENT,
					kw_list[k].kw[i+1] == 0 ? 10 : 4);
			lexa_add_tr(la, kw_list[k].sb+i-1,  'a', 'z',	NULL,	NULL,	S_IDENT,	T_IDENT	);
			lexa_add_tr(la, kw_list[k].sb+i-1,  'A', 'Z',	NULL,	NULL,	S_IDENT,	T_IDENT	);
			lexa_add_tr(la, kw_list[k].sb+i-1,  '0', '9',	NULL,	NULL,	S_IDENT,	T_IDENT	);
			lexa_add_tr(la, kw_list[k].sb+i-1,    0, 0,	NULL,	"_",	S_IDENT,	T_IDENT	);
		}
		lexa_add_tr(la, kw_list[k].sb+i-1,	  'a', 'z',	NULL,	NULL,	S_IDENT,	T_IDENT	);
		lexa_add_tr(la, kw_list[k].sb+i-1,	  'A', 'Z',	NULL,	NULL,	S_IDENT,	T_IDENT	);
		lexa_add_tr(la, kw_list[k].sb+i-1,	  '0', '9',	NULL,	NULL,	S_IDENT,	T_IDENT	);
		lexa_add_tr(la, kw_list[k].sb+i-1,	    0, 0,	NULL,	"_",	S_IDENT,	T_IDENT	);
		*/
	}

	/* operators */
	lexa_add_string(la, "+", T_OP_PLUS);
	lexa_add_string(la, "-", T_OP_MINUS);
	lexa_add_string(la, "*", T_OP_MUL);
	lexa_add_string(la, "/", T_OP_DIV_INT);
	lexa_add_string(la, ":=", T_OP_ASSIGN);
	lexa_add_string(la, "==", T_OP_EQ);
	lexa_add_string(la, ">", T_OP_GT);
	lexa_add_string(la, ">=", T_OP_GE);
	lexa_add_string(la, "<", T_OP_LT);
	lexa_add_string(la, "<=", T_OP_LE);
	lexa_add_string(la, "!=", T_OP_NEQ);
	lexa_add_string(la, "<>", T_OP_NEQ);
	/* brackets */
	lexa_add_string(la, "(", T_BRACKET_BEGIN);
	lexa_add_string(la, ")", T_BRACKET_END);
	/* delimiters */
	lexa_add_string(la, ";", T_SEMICOLON);
	lexa_add_string(la, ",", T_COMMA);
	lexa_add_string(la, ":", T_COLON);

#if DEBUG >= 1
	DBG_MSG(1, "State transition table:");
	lexa_table_dump(la->table, stderr);
#endif
	
	DBG_MSG(1, "Initializing Synta...");
	sa = synta_init(la);

	DBG_MSG(1, "Initializing complete.");

	if (argc == 2 && strcmp(argv[1], "lexa") == 0) {
		/*** LEXA TEST ***/
		/* read all data */
		DBG_MSG(1, "Processing data...");
		printf("*** LEXA ***\nReading file...\n");
		while ((token = lexa_scan(la, &type))) {
			printf("\tByte %6ld: Line %3ld: %16s: %s\n", lexa_get_byte(la), lexa_get_line(la),
					type >= 0 ? T_name[type]:"error",
					type != T_WHITESPACE ? token : "");
			fflush(stdout);
			free(token);
		}
		printf("Reading done (%s).\n", type >= 0 ? T_name[type] : "error");
		fflush(stdout);
	} else {
		/*** SYNTA TEST ***/
		DBG_MSG(1, "*** SYNTA ***\nProcessing data...");
		printf("Parsing file...\n");
		synta_parse(sa);
		printf("Parsing done.\n");
		fflush(stdout);
	}
	
	/* cleanup */
	DBG_MSG(1, "Cleanup...");
	lexa_cleanup(la);

	DBG_MSG(1, "End.");
	return 0;
}

