/*
 * synta.c
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdlib.h>
#include <assert.h>

#include "synta.h"
#include "lexa.h"
#include "tree.h"

#define DEBUG 10
#include "debug.h"
 
/****************************************************************************/

char *T_name[T_MAX + 1] = {
	"T_NONE", "T_WHITESPACE", "T_COMMENT",
	"T_BLOCK_BEGIN", "T_BLOCK_END", "T_BRACKET_BEGIN", "T_BRACKET_END",
	"T_COLON", "T_COMMA", "T_DOT", "T_END", "T_IDENT", "T_INT", "T_KW_DO",
	"T_KW_DOWNTO", "T_KW_ELSE", "T_KW_FOR", "T_KW_IF", "T_KW_INT",
	"T_KW_PROGRAM", "T_KW_REAL", "T_KW_THEN", "T_KW_TO", "T_KW_VAR",
	"T_KW_WHILE", "T_KW_WRITELN", "T_OP_ASSIGN", "T_OP_DIV",
	"T_OP_DIV_INT", "T_OP_EQ", "T_OP_GE", "T_OP_GT", "T_OP_LE", "T_OP_LT",
	"T_OP_MINUS", "T_OP_MOD", "T_OP_MUL", "T_OP_NEQ", "T_OP_PLUS",
	"T_REAL", "T_SEMICOLON",
	NULL
};

char *S_name[S_MAX + 1] = {
	"S_START", "S_WHITESPACE",
	"S_VARIABLE", "S_FUNCTION",
	"S_STRING_Q", "S_STRING_QS", "S_STRING_QT",
	"S_INTEGER", "S_INTEGER_0",
	"S_REAL",
	"S_COMMENT_C_B", "S_COMMENT_C", "S_COMMENT_C_E", "S_COMMENT_C_T",
	"S_COMMENT_CC", "S_COMMENT_CC_T",
	"S_COMMENT_S", "S_COMMENT_S_T",
	"S_IDENT",
	NULL
};

/****************************************************************************/

static int synta_read_next(SYNTA *synta, int expected_type);

#include "synta-gen.gc"

/****************************************************************************/

SYNTA *synta_init(LEXA *lexa) {
	SYNTA *synta;
	
	assert(lexa != NULL);

	synta = malloc(sizeof(SYNTA));
	if (synta) {
		synta->lexa = lexa;
		synta->tree = tree_new();
		synta->tree_iter = tree_iter_new(synta->tree);
		synta->token = NULL;
		synta->token_type = 0;
		return synta;
	} else {
		return NULL;
	}
}


void synta_cleanup(SYNTA *synta) {
	assert(synta != NULL);

	free(synta);
}

int synta_parse(SYNTA *synta) {
	assert(synta != NULL);

	DBG_MSG(1, "synta_parse: start");
	tree_append(synta->tree_iter);
	synta_read_next(synta, -1);
	synta_parse_start(synta);
	DBG_MSG(1, "synta_parse: end");

#if DEBUG >= 2
	tree_dump(synta->tree, stderr);
#endif

	return 0;
}

/****************************************************************************/

static int synta_read_next(SYNTA *synta, int expected_type)
	/* reads next token from lexa, ignore T_WHITESPACE and T_COMMENT
	 */
{
	assert(synta != NULL);
	assert(synta->lexa != NULL);

	if (expected_type != -1 && synta->token_type != expected_type) {
		fprintf(stderr, "%s:%ld: Unexpected token \"%s\". Aborting.",
				synta->lexa->file_name,
				lexa_get_line(synta->lexa),
				synta->token);
		exit(1);

	}

	if (expected_type != -1) {
		tree_update(synta->tree_iter, synta->token, synta->token_type);
		tree_append(synta->tree_iter);
	}
	
	
	/* read next token and ignore whitespace + comments*/
	do {
		synta->token = lexa_scan(synta->lexa, &(synta->token_type));
	} while (synta->token && (synta->token_type == T_WHITESPACE || synta->token_type == T_COMMENT));

	printf("\tByte %6ld: Line %3ld: %16s: %s\n", lexa_get_byte(synta->lexa), lexa_get_line(synta->lexa),
			synta->token_type >= 0 ? T_name[synta->token_type]:"error",
			synta->token_type != T_WHITESPACE ? synta->token : "");

	
	return synta->token != NULL;
}
	
