/*
 * synta.h
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#ifndef JK_SYNTA_H
#define JK_SYNTA_H

#include "lexa.h"
#include "tree.h"

typedef struct {
	LEXA *lexa;
	TREE *tree;
	TREE_ITER *tree_iter;
	
	char *token;
	int token_type;
} SYNTA;

SYNTA *synta_init(LEXA *lexa);
void synta_cleanup(SYNTA *synta);
int synta_parse(SYNTA *synta);

/*** Terminal symbols *******************************************************/

enum {
	T_ERR = -1,
	T_NONE = 0, T_WHITESPACE, T_COMMENT,
	T_KW_BEGIN, T_KW_END, T_BRACKET_BEGIN, T_BRACKET_END, T_COLON, T_COMMA,
	T_DOT, T_END, T_IDENT, T_INT, T_KW_DO, T_KW_DOWNTO, T_KW_ELSE,
	T_KW_FOR, T_KW_IF, T_KW_INT, T_KW_PROGRAM, T_KW_REAL, T_KW_THEN,
	T_KW_TO, T_KW_VAR, T_KW_WHILE, T_KW_WRITELN, T_OP_ASSIGN, T_OP_DIV,
	T_OP_DIV_INT, T_OP_EQ, T_OP_GE, T_OP_GT, T_OP_LE, T_OP_LT, T_OP_MINUS,
	T_OP_MOD, T_OP_MUL, T_OP_NEQ, T_OP_PLUS, T_REAL, T_SEMICOLON,
	T_MAX
};
extern char *T_name[T_MAX + 1];

enum {
	S_START, S_WHITESPACE,
	S_VARIABLE, S_FUNCTION,
	S_STRING_Q, S_STRING_QS, S_STRING_QT,
	S_INTEGER, S_INTEGER_0,
	S_REAL,
	S_COMMENT_C_B, S_COMMENT_C, S_COMMENT_C_E, S_COMMENT_C_T,
	S_COMMENT_CC, S_COMMENT_CC_T,
	S_COMMENT_S, S_COMMENT_S_T,
	S_IDENT,
	S_MAX,
};
extern char *S_name[S_MAX + 1];

/*** SYNTA ******************************************************************/


#endif

