/*
 * tree.h
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#ifndef JK_TREE_H
#define JK_TREE_H

typedef struct s_tree_node TREE_NODE;
struct s_tree_node {
	TREE_NODE *next, *prev, *parent, *child;

	int depth;
	
	/* data */
	char *token;
	int token_type;
};

typedef struct s_tree {
	TREE_NODE *root;	/* first in depth 0: parent = NULL, prev = NULL */
} TREE;

typedef struct s_tree_iter {
	TREE *tree;
	TREE_NODE *node;
} TREE_ITER;

/* create & destroy tree */
TREE *tree_new(void);
void tree_destroy(TREE* tree);

/* create & destroy single tree node */
TREE_NODE *tree_node_new(void);
void tree_node_destroy(TREE_NODE *node);

/* create & destroy iterator */
TREE_ITER *tree_iter_new(TREE *tree);
void tree_iter_destroy(TREE_ITER *iter);
TREE_ITER *tree_iter_dup(TREE_ITER *iter);

/* get data at iter */
const char *tree_get_token(TREE_ITER *iter);
int tree_get_token_type(TREE_ITER *iter);

/* update data at iter */
void tree_update(TREE_ITER *iter, const char *token, int token_type);

/* add node & move iter */
int tree_append(TREE_ITER *iter);
int tree_prepend(TREE_ITER *iter);
int tree_add_child(TREE_ITER *iter);

/* delete node and it's childs */
int tree_delete(TREE_ITER *iter);

/* iterator check */
int tree_iter_valid(TREE_ITER *iter);

/* iterator movement */
int tree_iter_move_next(TREE_ITER *iter);
int tree_iter_move_prev(TREE_ITER *iter);
int tree_iter_move_child(TREE_ITER *iter);
int tree_iter_move_parent(TREE_ITER *iter);
int tree_iter_move_next_linear(TREE_ITER *iter);

/* debug: dump entire tree to stream */
void tree_dump(TREE *tree, FILE *file);
void tree_dump_iter(TREE_ITER *iter, FILE *file, int depth);

#endif

